import codecs
import csv
import json
import logging
import os
import dateutil.parser
from _socket import gaierror

import requests
from util.exceptions import IllegalArgumentError
from util.regex import SO_URL_REGEX

# get root logger
logger = logging.getLogger('ghapi_logger')


class GhFile(object):
    """ Data structure for GH files """

    def __init__(self, id, url, line, link, repo_name, path):
        self.id = str(id).strip()
        self.url = str(url).strip()
        self.line = str(line).strip()
        self.link = str(link).strip()
        self.repo_name = str(repo_name).strip()
        self.path = str(path).strip()

        self.commit = GhCommit("", "")

    def __str__(self):
        return self.id + ", " + str(self.repo_name) + "/blob/master/" + str(self.path)


class GhCommit(object):
    """ Data structure for GH commits """

    def __init__(self, sha, date):
        self.sha = sha
        self.date = date
        self.diff_line = ""

    def __str__(self):
        return self.sha + ", " + self.date


class GhFileList(object):
    """ List with GH files """

    def __init__(self, name, start_index, chunk_size):
        self.name = name
        self.start_index = start_index
        assert start_index >= 0
        self.chunk_size = chunk_size
        self.list = []
        self.session = requests.Session()

    def read_from_csv(self, import_file, skip_header, delimiter):
        # see http://stackoverflow.com/a/844443
        with codecs.open(import_file, encoding='utf8') as fp:
            logger.info("Reading URLs from " + import_file + "...")
            reader = csv.reader(fp, delimiter=delimiter)
            if skip_header:
                next(reader, None)
            current_index = 0
            for row in reader:
                if current_index < self.start_index:
                    current_index += 1
                    continue
                elif current_index >= self.start_index+self.chunk_size:
                    current_index += 1
                    break
                if row:
                    if len(row) == 6:
                        id = str(row[0])
                        url = str(row[1])
                        line = str(row[2])
                        link = str(row[3])
                        repo_name = str(row[4])
                        path = str(row[5])
                        file_obj = GhFile(id, url, line, link, repo_name, path)
                        self.list.append(file_obj)
                    else:
                        raise IllegalArgumentError("Wrong CSV format.")
                current_index += 1
        logger.info(str(len(self.list)) + " URLs were imported.")

    def write_to_csv(self, export_dir, delimiter):
        if len(self.list) == 0:
            logger.info("Nothing to write...")
            return None

        if not os.path.exists(export_dir):
            os.makedirs(export_dir)

        file_path = os.path.join(
            export_dir,
            '{0}_{1}_{2}-{3}.csv'.format(self.name, "gh-data", str(self.start_index),
                                         str(self.start_index+min(len(self.list), self.chunk_size)-1))
        )

        # see http://stackoverflow.com/a/844443
        with codecs.open(file_path, 'w', encoding='utf8') as fp:
            logger.info('Writing file list to ' + file_path + '...')

            writer = csv.writer(fp, delimiter=delimiter)

            writer.writerow(["id", "url", "line", "link", "repo_name", "path",
                             "commit_sha", "commit_date", "commit_diff_line"])

            for i in range(len(self.list)):
                current = self.list[i]
                try:
                    writer.writerow([current.id, current.url, current.line, current.link, current.repo_name,
                                     current.path, current.commit.sha, str(current.commit.date),
                                     str(current.commit.diff_line)])
                except UnicodeEncodeError:
                    logger.error("Encoding error while writing data for url " + current.url)

            logger.info(str(len(self.list)) + ' URLs were exported.')

    def retrieve_gh_data(self, gh_access_token):
        logger.info("Retrieving data using GH API...")
        count = 0
        for current in self.list:

            commits = self.get_commits(current, gh_access_token)
            commit = self.search_so_url(current, commits, gh_access_token)

            if commit:
                count += 1
                current.commit = commit

        logger.info("Retrieved " + str(count) + " commits.")

    def search_so_url(self, gh_file, commits, gh_access_token):
        if commits:
            for commit in commits:
                url = "https://api.github.com/repos/" + gh_file.repo_name + "/commits/" + commit.sha\
                      + "?access_token=" + gh_access_token
                try:
                    commit_data = self.session.get(url)
                except (gaierror,
                        requests.exceptions.ConnectionError,
                        requests.packages.urllib3.exceptions.MaxRetryError,
                        requests.packages.urllib3.exceptions.NewConnectionError):
                    logger.error("An error occurred while getting commit " + str(commit.sha))

                if not commit_data:
                    return None

                commit_data_json = json.loads(commit_data.text)
                files = commit_data_json.get("files", None)

                if not files:
                    return None

                for file in files:
                    if file["filename"] == gh_file.path:
                        patch = file.get("patch", None)
                        if not patch:
                            return None
                        patch_lines = patch.split('\n')
                        for line in patch_lines:
                            if line.startswith("+") and line[1:].strip() == gh_file.line.strip():
                                commit.diff_line = line
                                return commit

        return None

    def get_commits(self, gh_file, gh_access_token):
        # this way, we don't find file that were renamed, moved, or merged after the SO reference was added
        # (request is limited to the specific repo and path)
        url = "https://api.github.com/repos/" + gh_file.repo_name + "/commits?path=" + gh_file.path\
              + "&access_token=" + gh_access_token
        try:
            commits = self.session.get(url)
        except (gaierror,
                requests.exceptions.ConnectionError,
                requests.packages.urllib3.exceptions.MaxRetryError,
                requests.packages.urllib3.exceptions.NewConnectionError):
            logger.error("An error occurred while getting commits for file with id " + str(gh_file.id))

        if commits:
            commit_list = []
            commits_json = json.loads(commits.text)
            for commit in commits_json:
                sha = commit["sha"]
                date = commit["commit"]["author"]["date"]
                # http://stackoverflow.com/a/3908349
                commit_list.append(GhCommit(sha, dateutil.parser.parse(date)))

            return sorted(commit_list, key=lambda c: c.date)  # oldest files first

        return None

    def __str__(self):
        result = ""

        for current_file in self.list:
            result += current_file + "\n"

        return result
