import codecs
import csv
import json
import logging
import os
import time
from _socket import gaierror
from bs4 import BeautifulSoup

import requests
from util.exceptions import IllegalArgumentError
from util.regex import SO_ANSWER_REGEX, SO_ID_REGEX, CODE_ELEMENT_REGEX, OUTER_CODE_BLOCK_REGEX
from util.regex import SO_QUESTION_REGEX

# get root logger
logger = logging.getLogger('soapi_logger')


class SoUrl(object):
    """ Data structure for SO answer or question """

    def __init__(self, url, line_count="", file_count="", language=""):
        self.url = str(url).strip()
        if line_count:
            self.line_count = int(str(line_count).strip())
        else:
            self.line_count = ""
        if file_count:
            self.file_count = int(str(file_count).strip())
        else:
            self.file_count = ""
        self.language = str(language).strip()
        self.inactive = False

        self.question_title = ""
        self.question_tags = ""
        self.question_has_accepted_answer = ""
        self.question_accepted_answer_id = ""
        self.question_accepted_answer_code_blocks = ""
        self.question_view_count = ""
        self.question_answer_count = ""

        self.question_id = ""
        self.question_score = ""
        self.question_share_link = ""
        self.question_comment_count = ""
        self.question_up_vote_count = ""
        self.question_down_vote_count = ""
        self.question_owner_reputation = ""
        self.question_editor_reputation = ""
        self.question_creation_date = ""
        self.question_last_edit_date = ""
        self.question_last_activity_date = ""

        self.answer_is_accepted = ""

        self.answer_id = ""
        self.answer_score = ""
        self.answer_share_link = ""
        self.answer_comment_count = ""
        self.answer_up_vote_count = ""
        self.answer_down_vote_count = ""
        self.answer_owner_reputation = ""
        self.answer_editor_reputation = ""
        self.answer_creation_date = ""
        self.answer_last_edit_date = ""
        self.answer_last_activity_date = ""
        self.answer_code_blocks = ""

    def __str__(self):
        return self.url + ", " + str(self.line_count) + " " + str(self.file_count)


class SoUrlList(object):
    """ List with SO answers and/or questions """

    def __init__(self, name, start_index, chunk_size, check_empty=False, check_code_block=False):
        self.name = name
        self.start_index = start_index
        assert start_index >= 0
        self.chunk_size = chunk_size
        self.check_empty = check_empty
        self.check_code_blocks = check_code_block
        self.list = []
        self.session = requests.Session()

    def read_from_csv(self, import_file, skip_header, delimiter):
        # see http://stackoverflow.com/a/844443
        with codecs.open(import_file, encoding='utf8') as fp:
            logger.info("Reading URLs from " + import_file + "...")
            reader = csv.reader(fp, delimiter=delimiter)
            if skip_header:
                next(reader, None)
            current_index = 0
            for row in reader:
                if current_index < self.start_index:
                    current_index += 1
                    continue
                elif current_index >= self.start_index+self.chunk_size:
                    current_index += 1
                    break
                if row:
                    if self.check_empty:
                        title = str(row[4])
                        language = str(row[0])
                        url = str(row[1])
                        line_count = int(row[2])
                        file_count = int(row[3])
                        url_obj = SoUrl(url, line_count, file_count, language)
                        url_obj.inactive = len(title) > 0
                        self.list.append(url_obj)
                    elif len(row) == 1:  # importing from answer sample drawn using R script
                        url = str(row[0])
                        url_obj = SoUrl(url)
                        self.list.append(url_obj)
                    elif len(row) == 3:
                        url = str(row[0])
                        line_count = int(row[1])
                        file_count = int(row[2])
                        url_obj = SoUrl(url, line_count, file_count)
                        self.list.append(url_obj)
                    elif len(row) == 4:
                        language = str(row[0])
                        url = str(row[1])
                        line_count = int(row[2])
                        file_count = int(row[3])
                        url_obj = SoUrl(url, line_count, file_count, language)
                        self.list.append(url_obj)
                    else:
                        raise IllegalArgumentError("Wrong CSV format.")
                current_index += 1
        logger.info(str(len(self.list)) + " URLs were imported.")

    def write_to_csv(self, export_dir, delimiter):
        if len(self.list) == 0:
            logger.info("Nothing to write...")
            return None

        if not os.path.exists(export_dir):
            os.makedirs(export_dir)

        if self.check_empty:
            suffix = "revised"
        else:
            suffix = "so-data"

        file_path = os.path.join(
            export_dir,
            '{0}_{1}_{2}-{3}.csv'.format(self.name, suffix, str(self.start_index),
                                         str(self.start_index+min(len(self.list), self.chunk_size)-1))
        )

        # see http://stackoverflow.com/a/844443
        with codecs.open(file_path, 'w', encoding='utf8') as fp:
            logger.info('Writing URL list to ' + file_path + '...')

            writer = csv.writer(fp, delimiter=delimiter)

            writer.writerow(["language", "url", "line_count", "file_count", "question_title", "question_tags",
                             "question_has_accepted_answer", "question_accepted_answer_id",
                             "question_accepted_answer_code_blocks",
                             "question_view_count", "question_answer_count",
                             "question_id", "question_score", "question_share_link", "question_comment_count",
                             "question_up_vote_count", "question_down_vote_count", "question_owner_reputation",
                             "question_editor_reputation", "question_creation_date", "question_last_edit_date",
                             "question_last_activity_date", "answer_is_accepted", "answer_id", "answer_score",
                             "answer_share_link", "answer_comment_count", "answer_up_vote_count",
                             "answer_down_vote_count", "answer_owner_reputation", "answer_editor_reputation",
                             "answer_creation_date", "answer_last_edit_date", "answer_last_activity_date",
                             "answer_code_blocks"])

            for i in range(len(self.list)):
                current = self.list[i]
                if self.check_empty and current.inactive:
                    continue
                try:
                    writer.writerow([current.language, current.url, current.line_count, current.file_count,
                                     current.question_title, current.question_tags,
                                     current.question_has_accepted_answer, current.question_accepted_answer_id,
                                     current.question_accepted_answer_code_blocks,
                                     current.question_view_count, current.question_answer_count, current.question_id,
                                     current.question_score, current.question_share_link, current.question_comment_count,
                                     current.question_up_vote_count, current.question_down_vote_count,
                                     current.question_owner_reputation, current.question_editor_reputation,
                                     current.question_creation_date, current.question_last_edit_date,
                                     current.question_last_activity_date, current.answer_is_accepted, current.answer_id,
                                     current.answer_score, current.answer_share_link, current.answer_comment_count,
                                     current.answer_up_vote_count, current.answer_down_vote_count,
                                     current.answer_owner_reputation, current.answer_editor_reputation,
                                     current.answer_creation_date, current.answer_last_edit_date,
                                     current.answer_last_activity_date, current.answer_code_blocks])
                except UnicodeEncodeError:
                    logger.error("Encoding error while writing data for url " + current.url)

            logger.info(str(len(self.list)) + ' URLs were exported.')

    def retrieve_so_data(self, so_key):
        logger.info("Setting custom filter for SO API...")
        try:
            response_filter = self.session.get('https://api.stackexchange.com/2.2/filters/create?include=answer.comment_count;'
                                'answer.down_vote_count;answer.last_editor;answer.link;answer.share_link;answer.tags;'
                                'answer.title;answer.up_vote_count;question.comment_count;question.down_vote_count;'
                                'question.favorite_count;question.last_editor;question.share_link;question.up_vote_count'
                                '&exclude=answer.community_owned_date;question.bounty_amount;question.bounty_closes_date;'
                                'question.community_owned_date;question.migrated_from;question.migrated_to&unsafe=false')
            so_filter = response_filter.json()["items"][0]["filter"]

            if self.check_code_blocks:
                response_filter = self.session.get(
                    'https://api.stackexchange.com/2.2/filters/create?include=answer.body&exclude=answer.tags;'
                    'answer.down_vote_count;answer.last_editor;answer.link;answer.share_link;answer.answer_id;'
                    'answer.community_owned_date;answer.creation_date;answer.is_accepted;answer.last_activity_date;'
                    'answer.last_edit_date;answer.locked_date;answer.owner;answer.question_id;answer.score;'
                    'answer.title;answer.up_vote_count&unsafe=false')
                so_body_filter = response_filter.json()["items"][0]["filter"]

        except (gaierror,
                requests.exceptions.ConnectionError,
                requests.packages.urllib3.exceptions.MaxRetryError,
                requests.packages.urllib3.exceptions.NewConnectionError):
            logger.error("An error occurred while setting the filter. Terminating...")
            exit(-1)
        except KeyError:
            logger.error("Empty response for filter request (daily quota reached?). Terminating...")
            exit(-1)

        logger.info("Received filter ID: " + so_filter)

        if self.check_empty:
            logger.info("Checking empty rows:")

        logger.info("Retrieving data using SO API...")
        count = 0
        for current in self.list:
            if self.check_empty and current.inactive:
                continue

            a_q_id = int(str(SO_ID_REGEX.findall(current.url)[0]).lower())

            if SO_ANSWER_REGEX.match(current.url):
                count += self.process_a(current, a_q_id, so_key, so_filter, so_body_filter, self.check_code_blocks)
                if current.question_id:
                    time.sleep(40/1000)  # sleep for 40ms due to SO API constraints
                    # don't check code blocks for corresponding question to save api calls
                    self.process_q(current, current.question_id, so_key, so_filter, False)

            elif SO_QUESTION_REGEX.match(current.url):
                count += self.process_q(current, a_q_id, so_key, so_filter, so_body_filter, self.check_code_blocks)

            time.sleep(40/1000)  # sleep for 40ms due to SO API constraints

        logger.info("Retrieved data for " + str(count) + " URLs.")

    def process_a(self, so_url, a_id, so_key, so_filter, so_body_filter="", check_code_blocks=False):
        response_a = self._get_a(a_id, so_key, so_filter)
        if response_a:
            response_a_json = json.loads(response_a.text)
            if len(response_a.json()["items"]) > 0:
                data = response_a_json["items"][0]

                so_url.answer_is_accepted = self._get_value(data, "is_accepted")
                so_url.answer_id = self._get_value(data, "answer_id")
                so_url.answer_score = self._get_value(data, "score")
                so_url.answer_share_link = self._get_value(data, "share_link")
                so_url.answer_comment_count = self._get_value(data, "comment_count")
                so_url.answer_up_vote_count = self._get_value(data, "up_vote_count")
                so_url.answer_down_vote_count = self._get_value(data, "down_vote_count")
                owner = self._get_value(data, "owner")
                if owner:
                    so_url.answer_owner_reputation = self._get_value(owner, "reputation")
                editor = self._get_value(data, "editor")
                if editor:
                    so_url.answer_editor_reputation = self._get_value(editor, "reputation")
                so_url.answer_creation_date = self._get_value(data, "creation_date")
                so_url.answer_last_edit_date = self._get_value(data, "last_edit_date")
                so_url.answer_last_activity_date = self._get_value(data, "last_activity_date")

                so_url.question_id = self._get_value(data, "question_id")

                if self.check_code_blocks:
                    time.sleep(40 / 1000)  # sleep for 40ms due to SO API constraints
                    so_url.answer_code_blocks = \
                        self._analyze_answer_code_blocks(so_url.answer_id, so_key, so_body_filter)

                return 1

        logger.error("Response was empty for answer with id " + str(a_id))
        return 0

    def process_q(self, so_url, q_id, so_key, so_filter, so_body_filter="", check_code_blocks=False):
        response_q = self._get_q(q_id, so_key, so_filter)
        if response_q:
            response_q_json = json.loads(response_q.text)

            if len(response_q.json()["items"]) > 0:
                data = response_q_json["items"][0]

                so_url.question_title = self._get_value(data, "title")
                so_url.question_tags = '; '.join(str(tag) for tag in self._get_value(data, "tags"))
                so_url.question_accepted_answer_id = self._get_value(data, "accepted_answer_id")
                so_url.question_has_accepted_answer = len(str(so_url.question_accepted_answer_id)) > 0

                so_url.question_view_count = self._get_value(data, "view_count")
                so_url.question_answer_count = self._get_value(data, "answer_count")

                so_url.question_id = self._get_value(data, "question_id")
                so_url.question_score = self._get_value(data, "score")
                so_url.question_share_link = self._get_value(data, "share_link")
                so_url.question_comment_count = self._get_value(data, "comment_count")
                so_url.question_up_vote_count = self._get_value(data, "up_vote_count")
                so_url.question_down_vote_count = self._get_value(data, "down_vote_count")
                owner = self._get_value(data, "owner")
                if owner:
                    so_url.question_owner_reputation = self._get_value(owner, "reputation")
                editor = self._get_value(data, "editor")
                if editor:
                    so_url.question_editor_reputation = self._get_value(editor, "reputation")
                so_url.question_creation_date = self._get_value(data, "creation_date")
                so_url.question_last_edit_date = self._get_value(data, "last_edit_date")
                so_url.question_last_activity_date = self._get_value(data, "last_activity_date")

                if check_code_blocks and so_url.question_has_accepted_answer:
                    time.sleep(40 / 1000)  # sleep for 40ms due to SO API constraints
                    so_url.question_accepted_answer_code_blocks =\
                        self._analyze_answer_code_blocks(so_url.question_accepted_answer_id, so_key, so_body_filter)

                return 1

        logger.error("Response was empty for question with id " + str(q_id))
        return 0

    def _get_value(self, json_list, index):
        # for some items such as data["editor"] or data["accepted_answer"] may not exist, there catch exception and return empty string
        try:
            return(json_list[index])
        except (IndexError, KeyError):
            return("")

    def _get_a(self, id, key, filter):
        try:
            return self.session.get(
                'https://api.stackexchange.com/2.2/answers/' + str(id) + "?key=" + key
                + "&site=stackoverflow&filter=" + filter)
        except (gaierror,
                requests.exceptions.ConnectionError,
                requests.packages.urllib3.exceptions.MaxRetryError,
                requests.packages.urllib3.exceptions.NewConnectionError):
            logger.error("An error occurred while getting data for answer with id " + str(id))

    def _get_q(self, id, key, filter):
        try:
            return self.session.get(
                'https://api.stackexchange.com/2.2/questions/' + str(id) + "?key=" + key
                + "&site=stackoverflow&filter=" + filter)
        except (gaierror,
                requests.exceptions.ConnectionError,
                requests.packages.urllib3.exceptions.MaxRetryError,
                requests.packages.urllib3.exceptions.NewConnectionError):
            logger.error("An error occurred while getting data for question with id " + str(id))

    def _analyze_answer_code_blocks(self, answer_id, key, body_filter):
        try:
            response = self.session.get(
                'https://api.stackexchange.com/2.2/answers/' + str(answer_id) + "?key=" + key
                + "&site=stackoverflow&filter=" + body_filter)
        except (gaierror,
                requests.exceptions.ConnectionError,
                requests.packages.urllib3.exceptions.MaxRetryError,
                requests.packages.urllib3.exceptions.NewConnectionError):
            logger.error("An error occurred while getting data for answer with id " + str(answer_id))

        if response:
            response_json = json.loads(response.text)
            if len(response.json()["items"]) > 0:
                data = response_json["items"][0]
                body = self._get_value(data, "body")

                html = BeautifulSoup(body, 'html.parser')
                code_blocks = html.find_all('code')

                count = len(code_blocks)
                length = 0
                result = ""

                for current in code_blocks:
                    content = current.get_text()
                    length += len(content)
                    summary = "//" + str(len(content))
                    result += summary

                return str(count) + "-" + str(length) + result
        else:
            return ""

    def __str__(self):
        result = ""

        for current_url in self.list:
            result += current_url + "\n"

        return result
